#!/usr/bin/env perl

# Copyright (C) 2009-2013  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.




# README
# This is the original version of svn-export. It is an ugly hackjob born of an
# itch and a single coding session and only mildy refined for general use on the
# second. It is deprecated in favor of the Python rewrite.

use strict;
use warnings;
use IPC::Open2;
use File::Basename;
use threads;
use Thread::Queue;
use Thread::Semaphore;

my $file_queue = Thread::Queue->new();
my $semaphore = Thread::Semaphore->new();



if (grep {$_ eq '--help'} @ARGV)
{
  print <<"HELP";
USAGE
Svn-export should work as a drop-in replacement for "svn export".

See "svn --help export" for standard svn export options.

SPECIFIC OPTIONS
      --help
  -r, --revision [<revision A>:]<revision B>
      --revision-file /path/to/file
      --threads n

See the man page for further details.

HELP
  exit;
}


my $SHELL='/bin/bash';
my ($shell_pipe_in,$shell_pipe_out,$shell_pid);
$shell_pid = open2($shell_pipe_out, $shell_pipe_in, $SHELL) or die "Unable to open pipes to $SHELL: $!\n";

my ($svn,$dir,$rev_a,$rev_b);
my $export_args = '';
my $global_args =  '';

my @export_args_0 = ('-q','--quiet','-N', '--non-recursive','--force','--ignore-externals','--ignore-keywords');
my @export_args_1 = ('--depth','--native-eol');
my @global_args_0 = ('--no-auth-cache','--non-interactive','--trust-server-cert');
my @global_args_1 = ('--username','--password','--config-dir','--config-option');
my @custom_args_1 = ('--revision-file','--threads');
my %custom_args;
@custom_args{@custom_args_1} = ('') x scalar @custom_args_1;


ARGV: for(my $i = 0;$i<scalar(@ARGV);$i++)
{
  my $argv = $ARGV[$i];
  foreach my $arg (@export_args_0)
  {
    if ($argv eq $arg)
    {
      $export_args .= "$arg ";
      next ARGV;
    }
  }
  foreach my $arg (@global_args_0)
  {
    if ($argv eq $arg)
    {
      $global_args .= "$arg ";
      next ARGV;
    }
  }
  foreach my $arg (@export_args_1)
  {
    if ($argv eq $arg)
    {
      $i++;
      $export_args .= "$argv ".$ARGV[$i].' ';
      next ARGV;
    }
  }
  foreach my $arg (@global_args_1)
  {
    if ($argv eq $arg)
    {
      $i++;
      $global_args .= "$argv ".$ARGV[$i].' ';
      next ARGV;
    }
  }
  foreach my $arg (@custom_args_1)
  {
    if ($argv eq $arg)
    {
      $i++;
      $custom_args{$arg} = $ARGV[$i];
      next ARGV;
    }
  }
  if ($argv eq '-r' or $argv eq '--revision')
  {
    $i++;
    my $revision = $ARGV[$i];
    if ($revision =~ m/^(.+):(.+)$/)
    {
      ($rev_a,$rev_b) = ($1,$2);
    }
    else
    {
      $rev_b = $revision;
    }
    next ARGV;
  }
  elsif (substr($argv,0,1) eq '-')
  {
    die "unknown argument: $argv\n";
  }
  elsif (not defined($svn))
  {
    $svn = $argv;
  }
  elsif (not defined($dir))
  {
    $dir = $argv;
  }
}

chop $export_args;
chop $global_args;

die "Please specify an svn repository.\n" if (not defined($svn));
$dir = basename($svn) if not defined($dir);
$dir .= '/' if (not substr($dir,-1) eq '/');


my ($rev) = (&shell("svn info $global_args $svn") =~ m/Revision:\s(\d+)/);
$rev_b = $rev if (not defined($rev_b) or uc($rev_b) eq 'HEAD');

if (defined($rev_a) and "$rev_a" ne "$rev_b")
{
  my @deleted = ();
  my @updated = ();
  foreach my $line (split "\n",&shell("svn diff -r $rev_a:$rev_b --summarize $global_args $svn"))
  {
    if ($line =~ m/^\s*(\w+)\s+(.+)\s*$/)
    {
      my ($change,$file) = ($1,$2);
      if (uc($change) eq 'D') {push @deleted,$file;}
      else {push @updated,$file;}
    }
  }
  foreach my $url (@deleted)
  {
    my $file_path = substr($url,length($svn));
    $file_path = substr($file_path,1) if (index($file_path,'/') == 1);
    my $localfile = $dir.$file_path;
    if (-e $localfile and -f $localfile)
    {
      print "removing $localfile\n";
      unlink $localfile;
    }
  }

  my $max_threads = ($custom_args{'--threads'} =~ m/(\d+)/) ? $1 : 30;
  $max_threads = 1 if ($max_threads < 1);
  my @threads;
  for (my $i=0;$i<$max_threads;$i++)
  {
    push @threads, threads->create(\&process_queue);
  }
  foreach my $url (@updated)
  {
    local $| = 1;
    my $file_path = substr($url,length($svn));
    $file_path = substr($file_path,1) if (index($file_path,'/') == 1);
    my $localfile = $dir.$file_path;
    my $file_dir = dirname($localfile);
    unlink $file_dir if (-e $file_dir and not -d $file_dir);
    &shell("mkdir -p $file_dir") if (not -e $file_dir);
    $file_queue->enqueue("${url}\@${rev_b} >$localfile");
  }
  $file_queue->enqueue(undef);
  foreach my $thr (@threads)
   {
     $thr->join();
   }
}
elsif (defined($rev_a) and "$rev_a" eq "$rev_b")
{
  print "local files are up-to-date\n";
}
else
{
  print "exporting revision $rev_b\n";
  &shell("svn export -r $rev_b --force $global_args $export_args $svn $dir",0,1);
}

&shell("find $dir -depth -type d -empty -exec rmdir {} \\;");

print "Local Exported Revision: $rev_b\n";
if (length($custom_args{'--revision-file'})>0)
{
  my $rev_file = $custom_args{'--revision-file'};
  open(my $fh,'>',$rev_file) or die "Unable to open $rev_file: $!\n";
  print $fh "$rev_b\n";
  close $fh;
}

&exitGracefully;


sub exitGracefully
{
  my ($exit_status) = (@_) ? @_ : 0;
  close $shell_pipe_in if defined($shell_pipe_in);
  waitpid ($shell_pid,0) if defined($shell_pid);
  exit($exit_status);
}

sub process_queue
{
  while (defined(my $cmd = $file_queue->dequeue()))
  {
    if ($cmd =~ m/^(.+)@(.+) >(.+)$/)
    {
      print "retrieving $1\n";
      system("svn cat $global_args $cmd");
    }
  }
  $file_queue->enqueue(undef);
}


sub shell
{
  my ($in,$return_exit_status,$print_lines) = @_;
  my $exit_status = 0;
  my $out = '';
  my $identifier = 'SHELL_PIPE_EOF';
  print $shell_pipe_in "$in\n";
  print $shell_pipe_in "echo \$?,\"$identifier\"\n";
  while (defined(my $line = <$shell_pipe_out>))
  {
    if ($line =~ m/^(\d+),\Q$identifier\E\n/)
    {
      $exit_status = $1;
      last;
    }
    else
    {
      $out .= $line;
      print $line if(defined($print_lines) and $print_lines);
    }
  }
  return (defined($return_exit_status) and $return_exit_status) ? ($out,$exit_status) : $out;
}
